<?php

namespace Drupal\multisite_manager_tool\Form;

use Drupal\Component\Utility\Crypt;
use Drupal\Component\Utility\UrlHelper;
use Drupal\Core\Database\Connection;
use Drupal\Core\Database\DatabaseExceptionWrapper;
use Drupal\Core\File\FileExists;
use Drupal\Core\File\FileSystem;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Site\Settings;
use Drupal\multisite_manager_tool\SitesHelper;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides an add new site form.
 */
final class AddNewSiteForm extends FormBase {

  /**
   * Sites helper service.
   *
   * @var \Drupal\multisite_manager_tool\SitesHelper
   */
  protected $sitesHelper;

  /**
   * File system service.
   *
   * @var \Drupal\Core\File\FileSystem
   */
  protected $fileSystem;

  /**
   * Database Service Object.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $database;

  /**
   * Settings Service Object.
   *
   * @var \Drupal\Core\Site\Settings
   */
  protected $settings;

  public function __construct(SitesHelper $sites_helper, FileSystem $file_system, Connection $database, Settings $settings) {
    $this->sitesHelper = $sites_helper;
    $this->fileSystem = $file_system;
    $this->database = $database;
    $this->settings = $settings;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    // Instantiates this form class.
    return new static(
      // Load the service required to construct this class.
      $container->get('sites_helper'),
      $container->get("file_system"),
      $container->get('database'),
      $container->get('settings')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId(): string {
    return 'multisite_manager_add_new_site_tool';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state): array {

    $form['help'] = [
      '#type' => 'item',
      '#title' => $this->t('Add new site process'),
      '#markup' => $this->t('<p>For the creation of the new site, the database connection details defined in database connection</p>
      <p>Make sure to configure the site URL in the virtual host or, if you are using DDEV, in the additional_hostnames section of the configuration file.</p>
      <p>The specified URL will be added to the main Drupal site configuration file.</p>'),
    ];

    $form['site_configuration'] = [
      '#type' => 'details',
      '#title' => $this->t('Site configuration'),
      '#open' => TRUE,
      '#weight' => 10,
    ];

    $form['site_configuration']['name'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Site name'),
      '#size' => 45,
      '#required' => TRUE,
    ];

    $form['site_configuration']['url'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Site url'),
      '#size' => 45,
      '#description' => $this->t("Enter a full URL, including 'http://' or 'https://'. Example: https://example.com."),
      '#required' => TRUE,
    ];

    $form['database_configuration'] = [
      '#type' => 'details',
      '#title' => $this->t('Database configuration'),
      '#weight' => 20,
    ];

    $form['database_configuration']['prefix'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Check it, if the application will be sharing default database, a unique table name prefix – such as drupal_ – will prevent collisions'),
    ];

    $form['database_configuration']['database'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Database name or prefix name'),
      '#size' => 45,
    ];

    $form['actions'] = ['#type' => 'actions'];
    $form['actions']['save'] = [
      '#type' => 'submit',
      '#value' => $this->t('Save'),
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state):void {
    $site_name = $form_state->getValue('name');
    $db_name = $form_state->getValue('database');
    $url = $form_state->getValue('url');
    $prefix = $form_state->getValue('prefix');

    if ($this->sitesHelper->checkIfExist($site_name)) {
      $form_state->setErrorByName('name', $this->t("Site name already exist, please try to choose another one"));
    }
    if (!UrlHelper::isValid($url, TRUE)) {
      $form_state->setErrorByName('url', $this->t("Invalid URL. Please enter a valid URL"));
    }

    $db_exists = $this->sitesHelper->checkIfDbExist($db_name);
    $db_prefix_exists = $this->sitesHelper->checkIfDbExist($db_name, TRUE);

    if (!$prefix && $db_exists) {
      $form_state->setErrorByName('database', $this->t("Database name already exists, please try to choose another one"));
    }
    elseif ($prefix && $db_prefix_exists) {
      $form_state->setErrorByName('prefix', $this->t("Prefix already exists, please try to choose another one"));
    }
    elseif (is_null($db_exists) || is_null($db_prefix_exists)) {
      $form_state->setErrorByName('database', $this->t("Something wrong happened with database connection. Please check the logs"));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state): void {
    if ($this->generateAliasingFeature($form_state->getValue('name'), $form_state->getValue('url')) && $this->generateSite($form_state->getValue('name'), $form_state->getValue('database'), $form_state->getValue('prefix'))) {
      $this->messenger()->addStatus($this->t("Site created successfully. Now you can start the drupal installation from the following link: <a href='@link_new_site' target='_blank'>@new_site</a>",
      [
        "@link_new_site" => $form_state->getValue('url'),
        "@new_site" => $form_state->getValue('name'),
      ]
      ));
    }
  }

  /**
   * Helper to generate site requirements.
   *
   * @param string $site_name
   *   The site name.
   * @param string $db_name
   *   The data base name.
   * @param bool $prefix
   *   Add prefix.
   *
   * @return bool
   *   If site could be created.
   */
  private function generateSite($site_name, $db_name = "", $prefix = FALSE):bool {
    $connection_options = $this->database->getConnectionOptions();
    if (!$prefix) {
      try {
        // Create new db.
        $db_name = $this->database->escapeLike($db_name);
        $username = $this->database->escapeLike($connection_options['username']);
        $this->database->query("CREATE DATABASE IF NOT EXISTS `$db_name` CHARACTER SET utf8mb4 COLLATE utf8mb4_general_ci;");
        $this->database->query("GRANT ALL PRIVILEGES ON `{$db_name}`.* TO :username@'localhost';", [
          ':username' => $username,
        ]);
        $this->database->query("GRANT ALL PRIVILEGES ON `{$db_name}`.* TO :username@'%';", [
          ':username' => $username,
        ]);
      }
      catch (DatabaseExceptionWrapper $e) {
        $this->messenger()->addError($this->t("Error creating new database: @e.", ["@e" => $e->getMessage()]));
        return FALSE;
      }
    }

    $directory = SitesHelper::SITES_DIRECTORY . $site_name;

    if (!$this->fileSystem->prepareDirectory($directory, FileSystemInterface::CREATE_DIRECTORY | FileSystemInterface::MODIFY_PERMISSIONS)) {
      $this->messenger()->addError($this->t("Error creating directory in: @path.", ["@path" => $directory]));
      return FALSE;
    }
    $default_file_settings_path = SitesHelper::SITES_DIRECTORY . "default/default.settings.php";
    if (!file_exists($this->fileSystem->realpath($default_file_settings_path))) {
      $this->messenger()->addError($this->t("File: @file does not exist", ["@file" => $default_file_settings_path]));
      return FALSE;
    }

    $new_file_settings_path = $directory . "/settings.php";
    if ($this->fileSystem->copy($default_file_settings_path, $new_file_settings_path, FileExists::Replace)) {
      $settings_content = file_get_contents($this->fileSystem->realpath($default_file_settings_path));
      $salt = (!$prefix) ? Crypt::randomBytesBase64(55) : $this->settings->get('hash_salt');

      $db = [
        "database" => ($prefix) ? $connection_options['database'] : $db_name,
        "username" => $connection_options["username"],
        "password" => $connection_options['password'],
        "host" => $connection_options['host'],
        "port" => $connection_options['port'],
        "driver" => $connection_options['driver'],
      ];

      $string_prefix = ($prefix) ? "'prefix' => '$db_name'" : '';
      $new_settings_content = $this->getHeredoc($site_name, "", $string_prefix, $db, $salt);

      $appended_content = $settings_content . $new_settings_content;
      if (!$this->fileSystem->saveData($appended_content, $new_file_settings_path, FileExists::Replace)) {
        $this->messenger()->addError($this->t("Error saving data in: @file", ["@file" => $$new_file_settings_path]));
        return FALSE;
      }
    }
    else {
      $this->messenger()->addError($this->t("Error copying file: @file to: @newfile",
        [
          "@file" => $default_file_settings_path,
          "@newfile" => $new_file_settings_path,
        ]));
      return FALSE;
    }
    return TRUE;
  }

  /**
   * Helper to generate directory aliasing feature.
   *
   * @param string $site_name
   *   The site name.
   * @param string $url
   *   The site url.
   *
   * @return bool
   *   If site alias could be created.
   */
  private function generateAliasingFeature($site_name, $url):bool {
    $sites_file = SitesHelper::SITES_DIRECTORY . "sites.php";
    if (!file_exists($this->fileSystem->realpath($sites_file))) {
      $this->messenger()->addError($this->t("File: @file does not exist", ["@file" => $sites_file]));
      return FALSE;
    }

    $parsed_url = parse_url($url);
    $host = $parsed_url['host'] ?? '';
    $port = isset($parsed_url['port']) ? $parsed_url['port'] . "." : '';
    $path = isset($parsed_url['path']) ? implode('.', explode('/', $parsed_url['path'])) : '';
    $url_site = $port . $host . $path;

    $sites_content = file_get_contents($this->fileSystem->realpath($sites_file));
    $new_sites_content = $this->getHeredoc($site_name, $url_site, "", [], "", "alias");

    $appended_content = $sites_content . $new_sites_content;
    if (!$this->fileSystem->saveData($appended_content, $sites_file, FileExists::Replace)) {
      $this->messenger()->addError($this->t("Error saving data in: @file", ["@file" => $sites_file]));
      return FALSE;
    }
    return TRUE;
  }

  /**
   * Helper to generate directory aliasing feature.
   *
   * @param string $site_name
   *   The site name.
   * @param string $url_site
   *   The site url.
   * @param string $string_prefix
   *   The database prefix.
   * @param array $db
   *   The db data.
   * @param string $salt
   *   The site sal.
   * @param string $op
   *   The string to get.
   *
   * @return bool
   *   If site could be created.
   */
  private function getHeredoc($site_name, $url_site, $string_prefix, $db, $salt, $op = "db"):string {
    if ($op === "alias") {
      return <<<PHP

\$sites['{$url_site}'] = '{$site_name}';
PHP;
    }

    return <<<PHP

\$settings['hash_salt'] = '{$salt}';

\$databases['default']['default'] = [
  'driver' => '{$db['driver']}',
  'database' => '{$db['database']}',
  'username' => '{$db['username']}',
  'password' => '{$db['password']}',
  'host' => '{$db['host']}',
  'port' => {$db['port']},
  {$string_prefix}
];

// Enable local development settings.
\$settings['container_yamls'][] = DRUPAL_ROOT . '/sites/development.services.yml';
\$settings['config_sync_directory'] = '../config/{$site_name}/sync';

// Set trusted host patterns.
\$settings['trusted_host_patterns'] = ['.*'];

PHP;
  }

}
